package com.simpledbbook;

import java.io.*;
import java.net.URL;
import java.util.*;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;

import com.simpledbbook.core.*;

public final class SimpleDB implements ISimpleDB {
  private static final String ALORITHM = "HmacSHA256";

  private final URL URL;
  private final Mac HMAC;
  private final String KEY;
  private final HTTPClient CLIENT;
  private ThreadLocal<String> lastToken = new ThreadLocal<String>();
  private ThreadLocal<String> lastSelect = new ThreadLocal<String>();

  public SimpleDB() {
    this("http://sdb.amazonaws.com/");
  }

  public SimpleDB(String serviceUrl) {
    try {
      Properties props = new Properties();
      String home = System.getProperty("user.home");
      props.load(new FileReader(new File(home, ".awssecret")));
      KEY = props.getProperty("AWS_ACCESS_KEY_ID");
      String secret = props.getProperty("AWS_SECRET_ACESSS_KEY");
      HMAC = Mac.getInstance(ALORITHM);
      HMAC.init(new SecretKeySpec(secret.getBytes(), ALORITHM));
      URL = new URL(serviceUrl);
      CLIENT = new HTTPClient();
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  public void createDomain(String domainName) {
    Request request = createNewRequest("CreateDomain");
    request.setDomain(domainName);
    CLIENT.fetch(request);
  }

  public void deleteDomain(String domainName) {
    Request request = createNewRequest("DeleteDomain");
    request.setDomain(domainName);
    CLIENT.fetch(request);
  }

  public List<String> listDomains() {
    Request request = createNewRequest("ListDomains");
    Response response = CLIENT.fetch(request);
    return response.extractDomains();
  }

  public Map<String, Long> domainMetadata(String domain) {
    Request request = createNewRequest("DomainMetadata");
    request.setDomain(domain);
    Response response = CLIENT.fetch(request);
    return response.extractDomainMetadata();
  }

  public void batchPut(String domain, List<Item> items) {
    Request request = createNewRequest("BatchPutAttributes");
    request.setDomain(domain);
    request.addItems(items);
    CLIENT.fetch(request);
  }

  public void put(String domain, Item item) {
    Request request = createNewRequest("PutAttributes");
    request.setItemName(item.getName());
    request.addAttributes(item.getAtts());
    request.setDomain(domain);
    CLIENT.fetch(request);
  }

  public void putIfNotExists(String domain, Item item,
      String expectedName) {
    Request request = createNewRequest("PutAttributes");
    request.setDomain(domain);
    request.setItemName(item.getName());
    request.addAttributes(item.getAtts());
    request.setExpectedToNotExist(expectedName);
    CLIENT.fetch(request);
  }

  public void putIfExists(String domain, Item item,
      String expectedName, String expectedValue) {
    Request request = createNewRequest("PutAttributes");
    request.setDomain(domain);
    request.setItemName(item.getName());
    request.addAttributes(item.getAtts());
    request.setExpectedAttribute(expectedName, expectedValue);
    Response response = CLIENT.fetch(request);
    if (response.extractErrorCode().startsWith("Condition")) {
      throw new RuntimeException(response.getBody());
    }
  }

  public void delete(String domain, Item item) {
    Request request = createNewRequest("DeleteAttributes");
    request.setDomain(domain);
    request.setItemName(item.getName());
    request.addAttributes(item.getAtts());
    CLIENT.fetch(request);
  }

  public void deleteIfNotExists(String domain, Item item,
      String expectedName) {
    Request request = createNewRequest("DeleteAttributes");
    request.setDomain(domain);
    request.setItemName(item.getName());
    request.addAttributes(item.getAtts());
    request.setExpectedToNotExist(expectedName);
    CLIENT.fetch(request);
  }

  public void deleteIfExists(String domain, Item item,
      String expectedName, String expectedValue) {
    Request request = createNewRequest("DeleteAttributes");
    request.setDomain(domain);
    request.setItemName(item.getName());
    request.addAttributes(item.getAtts());
    request.setExpectedAttribute(expectedName, expectedValue);
    CLIENT.fetch(request);
  }

  public Item get(String domain, String itemName, String... atts) {
    Request request = createNewRequest("GetAttributes");
    request.setDomain(domain);
    request.setItemName(itemName);
    request.addAttributeNames(atts);
    Response response = CLIENT.fetch(request);
    return new Item(itemName, response.extractAttributes());
  }

  public Item getConsistent(String domain, String itemName,
      String... atts) {
    Request request = createNewRequest("GetAttributes");
    request.setDomain(domain);
    request.setItemName(itemName);
    request.addAttributeNames(atts);
    request.setConsistentRead();
    Response response = CLIENT.fetch(request);
    return new Item(itemName, response.extractAttributes());
  }

  public List<Item> select(String select) {
    return select(select, "");
  }

  public List<Item> selectConsistent(String select) {
    return selectConsistent(select, "");
  }

  public boolean hasNextPage() {
    return lastToken.get().length() > 0;
  }

  public List<Item> getNextPage() {
    return select(lastSelect.get(), lastToken.get());
  }

  public List<Item> getNextPageConsistent() {
    return selectConsistent(lastSelect.get(), lastToken.get());
  }

  private List<Item> select(String select, String nextToken) {
    lastSelect.set(select);
    Request request = createNewRequest("Select");
    request.setSelectExpression(select);
    request.setNextToken(nextToken);
    Response response = CLIENT.fetch(request);
    lastToken.set(response.extractNextToken());
    return response.extractItems();
  }

  private List<Item> selectConsistent(String select, String token) {
    lastSelect.set(select);
    Request request = createNewRequest("Select");
    request.setSelectExpression(select);
    request.setNextToken(token);
    request.setConsistentRead();
    Response response = CLIENT.fetch(request);
    lastToken.set(response.extractNextToken());
    return response.extractItems();
  }

  private Request createNewRequest(String action) {
    return new Request(HMAC, action, KEY, URL);
  }

  public long count(String domain) {
    return count(domain, "");
  }

  public int count(String domain, String whereClause) {
    return count(domain, whereClause, Integer.MAX_VALUE);
  }

  public int count(String domain, String where, int limit) {
    String query = "SELECT count(*) FROM `" + domain + "` " + where;
    List<Item> items = select(query);
    int count = 0;
    count = extractCount(items);
    while (hasNextPage() && count < limit) {
      items = getNextPage();
      count += extractCount(items);
    }
    return count;
  }

  private int extractCount(List<Item> items) {
    int result = 0;
    if (items.size() > 0) {
      Item item = items.get(0);
      result = Integer.parseInt(item.getAttribute("Count"));
    }
    return result;
  }

  public List<Item> selectWithRealLimit(String select, int limit) {
    List<Item> result = select(select);
    while (result.size() < limit && hasNextPage()) {
      result.addAll(getNextPage());
    }
    return result;
  }
}
